import torch
import torch.nn as nn
from .base_forecaster import BaseForecaster

class TransformerForecaster(BaseForecaster):
    def __init__(self, input_dim, model_dim, num_layers, nhead, forecast_horizon):
        super(TransformerForecaster, self).__init__()
        self.encoder = nn.Linear(input_dim, model_dim)
        decoder_layer = nn.TransformerDecoderLayer(
            d_model=model_dim, nhead=nhead, dim_feedforward=model_dim*4, dropout=0.1
        )
        self.decoder = nn.TransformerDecoder(decoder_layer, num_layers=num_layers)
        self.positional_encoding = nn.Parameter(self._generate_positional_encoding(forecast_horizon, model_dim), requires_grad=False)
        self.fc_out = nn.Linear(model_dim, input_dim)
        self.forecast_horizon = forecast_horizon

    def _generate_positional_encoding(self, forecast_horizon, model_dim):
        pe = torch.zeros(forecast_horizon, model_dim)
        position = torch.arange(0, forecast_horizon, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, model_dim, 2).float() * (-torch.log(torch.tensor(10000.0)) / model_dim))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        return pe

    def forward(self, x):
        batch_size, T_in, input_dim = x.size()
        memory = self.encoder(x).transpose(0, 1)
        tgt = self.positional_encoding.unsqueeze(1).repeat(1, batch_size, 1)
        out = self.decoder(tgt, memory)
        preds = self.fc_out(out).transpose(0, 1)
        return preds

    def train_model(self, train_loader, num_epochs, learning_rate, device):
        self.to(device)
        optimizer = torch.optim.Adam(self.parameters(), lr=learning_rate)
        criterion = nn.MSELoss()
        for epoch in range(num_epochs):
            self.train()
            epoch_loss = 0.0
            for x_batch, y_batch in train_loader:
                x_batch = x_batch.to(device).squeeze(1)
                y_batch = y_batch.to(device).transpose(1,2)
                optimizer.zero_grad()
                preds = self(x_batch)
                loss = criterion(preds, y_batch)
                loss.backward()
                optimizer.step()
                epoch_loss += loss.item() * x_batch.size(0)
            print(f"Epoch {epoch+1}/{num_epochs}, Loss: {epoch_loss/len(train_loader.dataset):.4f}")
        return

    def predict(self, new_x, forecast_horizon, device):
        self.eval()
        new_x = new_x.to(device).squeeze(1)
        # If new_x is 2D, add the feature dimension.
        if new_x.dim() == 2:
            new_x = new_x.unsqueeze(-1)  # Now shape is (batch, T_in, 1)
        with torch.no_grad():
            preds = self(new_x)
        return preds, None


    def evaluate(self, test_loader, device):
        self.eval()
        mse_total, mae_total, count = 0.0, 0.0, 0
        with torch.no_grad():
            for x_batch, y_batch in test_loader:
                x_batch = x_batch.to(device).squeeze(1)
                y_batch = y_batch.to(device).squeeze(1) 
                preds = self(x_batch)
                mse = ((preds.squeeze(-1) - y_batch) ** 2).mean().item()
                mae = (torch.abs(preds.squeeze(-1) - y_batch)).mean().item()
                mse_total += mse * x_batch.size(0)
                mae_total += mae * x_batch.size(0)
                count += x_batch.size(0)
        mse_avg = mse_total / count
        mae_avg = mae_total / count
        print(f"Evaluation -> MSE: {mse_avg:.4f}, MAE: {mae_avg:.4f}")
        return mse_avg, mae_avg



import torch
import torch.nn as nn
from torchdiffeq import odeint_adjoint

class HierarchicalNeuralDynamics(nn.Module):
    def __init__(self, hidden_dim, input_dim):
        super(HierarchicalNeuralDynamics, self).__init__()
        self.global_net = nn.Sequential(
            nn.Linear(hidden_dim + input_dim + 1, 64),
            nn.ReLU(),
            nn.Linear(64, hidden_dim * 2)
        )
        self.local_net = nn.Sequential(
            nn.Linear(hidden_dim + input_dim + 1, 64),
            nn.ReLU(),
            nn.Linear(64, hidden_dim * 2)
        )
    
    def forward(self, t, state, level='global'):
        h, x_scale = state 
        t_scalar = t if t.dim() == 0 else t[0]
        time_feature = t_scalar.unsqueeze(0).expand(h.size(0), 1)
        inp = torch.cat([h, x_scale, time_feature], dim=-1)
        if level == 'global':
            update = self.global_net(inp)
        else:
            update = self.local_net(inp)
        delta_mu, delta_logvar = torch.chunk(update, 2, dim=-1)
        return delta_mu, delta_logvar

def adaptive_hierarchical_ode_solver(dynamics, h0, logvar0, t_span, x_scale_sequence):
    """
    Solves the ODE for hierarchical dynamics.
    Currently only uses the 'global' level.
    """
    class ODEFunc(nn.Module):
        def __init__(self, dynamics, x_sequence, t_span, level):
            super(ODEFunc, self).__init__()
            self.dynamics = dynamics
            self.x_sequence = x_sequence 
            self.t_span = t_span
            self.level = level
            
        def forward(self, t, state):
            h, logvar = state
            idx = (torch.abs(self.t_span - t)).argmin()
            x_t = self.x_sequence[:, idx, :] 
            return self.dynamics(t, (h, x_t), level=self.level)
    
    global_func = ODEFunc(dynamics, x_scale_sequence, t_span, level='global')
    state0 = (h0, logvar0)
    sol_global = odeint_adjoint(global_func, state0, t_span, rtol=1e-3, atol=1e-3)
    h_sol = sol_global[0].transpose(0, 1) 
    logvar_sol = sol_global[1].transpose(0, 1) 
    return h_sol, logvar_sol



import torch
import numpy as np

def jitter(x, sigma=0.03):
    """Add random Gaussian noise."""
    return x + sigma * torch.randn_like(x)

def scaling(x, sigma=0.1):
    """Multiply by a random scaling factor."""
    factor = torch.randn(x.size(0), 1, 1, device=x.device) * sigma + 1.0
    return x * factor

def time_warp(x, max_warp=0.2):
    """Time warp via re–interpolation of the series."""
    batch_size, channels, length = x.size()
    warp = torch.linspace(0, 1, steps=length, device=x.device) + (torch.rand(length, device=x.device) - 0.5) * max_warp
    warp, _ = torch.sort(warp)
    orig_idx = torch.linspace(0, 1, steps=length, device=x.device)
    x_warped = torch.zeros_like(x)
    for b in range(batch_size):
        for c in range(channels):
            x_np = x[b, c, :].cpu().numpy()
            warp_np = warp.cpu().numpy()
            orig_np = orig_idx.cpu().numpy()
            interp_np = np.interp(orig_np, warp_np, x_np)
            x_warped[b, c, :] = torch.tensor(interp_np, device=x.device)
    return x_warped



import torch
import torch.nn as nn

class TimeAwareTransformerDecoder(nn.Module):
    def __init__(self, hidden_dim, output_dim, forecast_horizon, num_layers=1, nhead=4):
        super(TimeAwareTransformerDecoder, self).__init__()
        self.forecast_horizon = forecast_horizon
        self.output_dim = output_dim
        self.positional_encoding = self._generate_positional_encoding(forecast_horizon, hidden_dim)
        
        decoder_layer = nn.TransformerDecoderLayer(
            d_model=hidden_dim, nhead=nhead, dim_feedforward=hidden_dim * 4, dropout=0.1
        )
        self.transformer_decoder = nn.TransformerDecoder(decoder_layer, num_layers=num_layers)
        self.fc_out = nn.Linear(hidden_dim, output_dim)
        self.fc_logvar = nn.Linear(hidden_dim, output_dim)
    
    def _generate_positional_encoding(self, forecast_horizon, hidden_dim):
        pe = torch.zeros(forecast_horizon, hidden_dim)
        position = torch.arange(0, forecast_horizon, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, hidden_dim, 2).float() * (-torch.log(torch.tensor(10000.0)) / hidden_dim))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        return nn.Parameter(pe, requires_grad=False)
    
    def forward(self, hidden):
        batch_size = hidden.size(1)
        tgt = self.positional_encoding.unsqueeze(1).repeat(1, batch_size, 1)
        out = self.transformer_decoder(tgt, hidden)
        preds = self.fc_out(out)
        pred_logvars = self.fc_logvar(out)
        return preds.transpose(0, 1), pred_logvars.transpose(0, 1)



import torch.nn as nn
from .dynamic_dilation_conv import DynamicDilationConv

class ResidualMultiScaleDecomposer(nn.Module):
    def __init__(self, num_scales, input_channels, filter_size):
        super(ResidualMultiScaleDecomposer, self).__init__()
        self.num_scales = num_scales
        self.paths = nn.ModuleList()
        for _ in range(num_scales):
            self.paths.append(nn.Sequential(
                DynamicDilationConv(input_channels, input_channels, filter_size),
                nn.BatchNorm1d(input_channels),
                nn.ReLU(),
                nn.Conv1d(input_channels, 1, filter_size, padding=filter_size // 2),
                nn.BatchNorm1d(1),
                nn.ReLU()
            ))
    
    def forward(self, x):
        return [path(x) for path in self.paths]



from .base_forecaster import BaseForecaster
from .apdtflow import APDTFlow
from .transformer_forecaster import TransformerForecaster
from .tcn_forecaster import TCNForecaster
from .ensemble_forecaster import EnsembleForecaster



# from setuptools import setup, find_packages

# setup(
#     name='apdtflow',
#     version='0.1.3',
#     packages=find_packages(),
#     install_requires=[
#         'torch',
#         'torchdiffeq',
#         'pandas',
#         'numpy',
#         'matplotlib',
#         'tensorboard',
#         'pyyaml',
#     ],
#     description='APDTFlow: A modular forecasting framework for time series data',
#     author='Yotam Barun',
#     author_email='yotambarun93@gmail.com',
#     url='https://github.com/yotambraun/APDTFlow',  # Main homepage URL
#     project_urls={
#         'Homepage': 'https://github.com/yotambraun/APDTFlow',
#         'Documentation': 'https://github.com/yotambraun/APDTFlow',
#         'Source': 'https://github.com/yotambraun/APDTFlow',
#     },
# )
from setuptools import setup, find_packages
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='apdtflow',
    version='0.1.4',
    packages=find_packages(),
    install_requires=[
        'torch',
        'torchdiffeq',
        'pandas',
        'numpy',
        'matplotlib',
        'tensorboard',
        'pyyaml',
    ],
    description='APDTFlow: A modular forecasting framework for time series data',
    long_description=long_description, 
    long_description_content_type="text/markdown", 
    author='Yotam Barun',
    author_email='yotambarun93@gmail.com',
    url='https://github.com/yotambraun/APDTFlow',
    project_urls={
        'Homepage': 'https://github.com/yotambraun/APDTFlow',
        'Documentation': 'https://github.com/yotambraun/APDTFlow',
        'Source': 'https://github.com/yotambraun/APDTFlow',
    },
)

